"""
The Markov Class:

Markov(data, periods, z, pprior = None, esupport = 3, eprior = None)

Based on the conditional Markov model developed in Golan (2008: 116), the Markov 
class estimates non-stationary transition probability matrices (TPMs) via 
generalized maximum entropy (GME) or generalized cross entropy (GCE). The class 
incorporates heteroskedasticity in the error structure, provides standard errors 
for the elements in the TPM, calculates some basic diagnostic statistics, and 
provides state elasticities for examination of the impact of the conditioning 
variables on the TPM (further discussed below).

The data is provided to the Markov class as an NT-by-K array (which is sorted on 
t then i) where i=1,...,N is the index for the observations in each 
cross-section of the panel, t=1,...,T is the time period index (Note: T or 
periods is also passed to the class), and K is the number of states used in the 
model. The conditioning variables are provided as an NT-by-S array z (again 
sorted on t then i) where S is the number of conditioning variable used. As for 
prior information on the TPM, it must be passed to pprior as a K-by-K array 
(Note: Priors should have no zero values. If zero values are desired use 
1e-05.). If no prior is provided, the uniform prior is used (i.e. the model 
defaults to GME formalism). The model allows for the use of 3 or 5 support 
points, but defaults to the use of 3 support points. The choice of support 
points is passed to esupport. Note that the class calculates the actual support 
vector by assuming the state data is generated by a Poisson process and then 
calculating the vector in accordance with Golan et al. (2007). Finally, the 
choice of priors for the error term is passed to eprior as an N(T-1)K-by-M array 
where M is the number of support points chosen. The rows of these priors must be 
sorted so as to "cycle" through i, then t, then j. If no priors for the error 
term are specified, the model defaults to a uniform prior.

Standard errors for the TPM are calculated by providing the self.params argument 
(i.e. the Lagrange multipliers) to self.se(). A number of inferential statistics 
are available via the diagnostics and yelasticity functions. The model's 
entropy, normalized entropy, information index, the entropy ratio statistic as 
well as the associated p-value are given by calling self.diagnostics() and 
printing self.entropy, self.norment, self.infoidx, self.entratio, or self.pval. 
The normalized entropy measure and information index are also available for each 
row of the TPM and can be called on by printing self.normentk or self.infoidxk. 
Lastly, the elasticity of the proportion of firms in each state with respect to 
each conditioning variable can be viewed by calling on self.yelasticity() and 
printing self.yelasticity, which provides an S-by-K matrix of elasticities (see 
Tonini and Jongeneel [2009] for the calculations).

"""
# Import relevant packages and basic set-up
import numpy as np
import matplotlib.pyplot as plt
from scipy import optimize
from scikits.statsmodels.sandbox.maxentropy.infotheo import logsumexp
from scikits.statsmodels.sandbox.regression.numdiff import approx_fprime_cs
from scipy.stats.stats import chisqprob
np.set_printoptions(precision = 3, linewidth=160, suppress = True)

class Markov(object):
    """A class for estimation of conditional Markov models via Maximum Entropy.

    Methods
    -------
    dual : defines dual formulation
    pmf : calculates P and W matrices
    gradient : defines gradient of dual formulation
    fit : conducts numerical optimization of the dual formulation
    se : calculates standard errors
    diagnostics : calculates diagnostic statistics
    yelasticity : calculates state elasticities

    """
    def __init__(self, data, periods, z, pprior = None, esupport = 3,
        eprior = None):                 
        """Initialize Markov class.

        Parameters
        ----------
        data : array
            Transition data
        periods : float
            Number of time periods
        z : array
            Conditioning data
        pprior : array
            Prior for transition probability matrix
        esupport : float
            Number of points in support vector
        eprior : array
            Prior for support weights

        """
        # Set-up state data
        self.data = np.asarray(data)
        K = self.data.shape[1]          # Get K from shape of state data
        self.K = K
        T = periods                     # T is time periods NOT transitions 
        self.T = T
        N = self.data.shape[0]/T        # N is number of obs.
        self.N = N
        x = self.data[: -N]                     
        self.x = x                      # Rhs data goes from t=1 to T-1
        self.X = np.kron(np.eye(K), x)  # Array is N(T-1)K-by-K^2
        y = self.data[N: ]              # Lhs data goes from t=2 to T
        self.Y = y.ravel('F')[:, None]  # Array is N(T-1)K-by-1
        # Set-up Z data
        self.z = np.asarray(z)          # Array is NT-by-S
        self.S = self.z.shape[1]
        self.zrhs = self.z[: -N]        # Rhs data goes from t=1 to T-1
        self.zlhs = self.z[N: ]         # Lhs data goes from t=2 to T
        self.Zrhs = np.kron(np.eye(K), self.zrhs) # Array is N(T-1)K-by-SK
        self.Zlhs = np.kron(np.eye(K), self.zlhs) # Array is N(T-1)K-by-SK
        # Set-up priors and support vectors
        if pprior == None:              # Default to uniform priors
            pprior = np.ones((K,K))/float(K)
        self.pprior = pprior.ravel('F')[:, None]  # Array is K^2-by-1
        if esupport == 3:               # V is K-by-M (below)
            self.v = np.array([[-1, 0, 1]])
        if esupport == 5:
            self.v = np.array([[-1,-0.5, 0, 0.5, 1]]) 
        self.V = np.kron(3 * np.sqrt(np.sum(y, axis=0))[:, None], self.v) / \
            float(np.amax(3 * np.sqrt(np.sum(y, axis=0))))
        if eprior == None:              # Default to uniform priors
            eprior = np.ones((N * (T-1) * K, np.size(self.v))) / \
            float(np.size(self.v))
        self.eprior = eprior            # Array is N(T-1)K-by-M
         
    def dual(self, params):   
        """Define dual formulation.

        Parameters
        ----------
        params : array
            Lagrange multipliers

        """  
        N = self.N   
        K = self.K
        T = self.T
        Y = self.Y                           
        X = self.X                                                   
        V = self.V
        Zrhs = self.Zrhs
        Zlhs = self.Zlhs
        pprior = self.pprior
        eprior = self.eprior
        sumomega = np.sum(logsumexp((np.log(pprior) + \
            np.dot(np.dot(X.T, Zrhs), \
            params[:, None])).reshape(K, K).T, axis = 1))
        sumpsi = np.sum(logsumexp(np.log(eprior) + \
            np.repeat(V, N * (T-1),axis = 0) * \
            np.tile(np.dot(Zrhs, params[:, None]), V.shape[1]),axis = 1))
        return (np.dot(np.dot(Y.T, Zlhs), params[:, None]) - \
            sumomega - sumpsi).item() 
             
    def pmf(self, params):  
        """Calculate P and W matrices.

        Parameters
        ----------
        params: array
            Lagrange multipliers

        """
        N = self.N                                 
        K = self.K                              
        T = self.T
        X = self.X                                  
        V = self.V
        Zrhs = self.Zrhs
        pprior = self.pprior                    
        eprior = self.eprior
        logpprob = np.log(pprior) + np.dot(np.dot(X.T,Zrhs),params[:, None]) - \
            np.tile(logsumexp((np.log(pprior) + np.dot(np.dot(X.T,Zrhs), \
            params[:, None])).reshape(K, K).T, axis = 1)[:, None], (K, 1))
        pprob = np.exp(logpprob)        # Array is K^2-by-1
        logwprob = np.log(eprior) + np.repeat(V, N * (T-1), axis = 0) * \
            np.tile(np.dot(Zrhs, params[:, None]), V.shape[1]) - \
            np.tile(logsumexp(np.log(eprior) + \
            np.repeat(V, N * (T-1), axis = 0) * \
            np.tile(np.dot(Zrhs, params[:, None]), V.shape[1]), axis = 1) \
            [:, None],V.shape[1])
        wprob = np.exp(logwprob)        # Array is N(T-1)K-by-M
        self.P = np.reshape(pprob,(K,K)).T   
        self.W = wprob
        return np.squeeze(pprob), np.squeeze(wprob)
        
    def gradient(self, params):   
        """Define gradient of dual formulation.

        Parameters
        ----------
        params: array
            Lagrange multipliers

        """
        N = self.N 
        T = self.T
        Y = self.Y                                 
        X = self.X                          
        V=  self.V
        Zrhs = self.Zrhs
        Zlhs = self.Zlhs
        pprob,wprob = self.pmf(params)
        return np.squeeze(np.dot(Zlhs.T,Y)) - np.dot(np.dot(Zrhs.T,X),pprob) - \
            np.dot(Zrhs.T,np.sum(np.repeat(V, N * (T-1), axis = 0) * \
            wprob, axis = 1))
        
    def fit(self, gtol = None):  
        """Conduct numerical optimization of the dual formulation.

        Parameters
        ----------
        gtol : float
            Gradient tolerance

        """
        K = self.K                              
        S = self.S
        params = np.zeros(S * K)  # There are SK params
        func = lambda x : -self.dual(x)
        funcprime = lambda x: -self.gradient(x)
        if gtol == None:
            gtol = 1e-04
        results = optimize.fmin_bfgs(func, params, funcprime, gtol = gtol,
            maxiter = 2400)
        self.params = results

    def se(self, params):        
        """Calculate standard errors.

        Parameters
        ----------
        params: array
            Lagrange multipliers

        """
        K = self.K 
        func1 = lambda x : self.gradient(x)
        hess = approx_fprime_cs(params, func1)
        iml = - hess             # Info matrix for lambdas        
        func2 = lambda x : self.pmf(x)[0]
        pprime = approx_fprime_cs(params, func2)
        impinv = np.dot(pprime, np.dot(np.linalg.pinv(iml), pprime.T)) 
        self.se = np.sqrt(impinv.diagonal()[:, None].reshape(K,K)).T

    def diagnostics(self):      
        """Calculate diagnostic statistics."""
        K = self.K
        S = self.S
        P = self.P + 1e-06      # Both P and P0 should be K-by-K
        P0 = np.reshape(self.pprior, (K, K)).T + 1e-06
        W = self.W + 1e-06      # Both W and W0 are N(T-1)K-by-M
        W0 = self.eprior
        HO = - np.sum(P0 * np.log(P0))         
        HA = - np.sum(P * np.log(P))         
        self.entropy = HA - np.sum(W * np.log(W))
        HOk = - np.sum(P0 * np.log(P0), axis = 1)[:, None]
        HAk = - np.sum(P * np.log(P), axis = 1)[:, None]
        self.norment = HA / HO                 # Normalized entropy (TPM)
        self.normentk = HAk / HOk              # Kth normalized entropy (TPM)
        self.infoidx = 1 - self.norment        # Information index (TPM)
        self.infoidxk = 1 - self.normentk      # Kth information index (TPM)       
        self.entratio = 2 * (HO - np.sum(W0 * np.log(W0))) - \
            2 * (HA - np.sum(W * np.log(W)))   # Entropy ratio    
        self.pval = chisqprob(self.entratio, S * K - 1)
                
    def yelasticity(self):    
        """Calculate state elasticities."""
        K = self.K              
        S = self.S
        N = self.N
        T = self.T  
        zrhs = self.z[: -N]             
        params = self.params[:, None]        
        pprob = self.P
        x = self.data[: -N]  
        y = self.data[N: ]                                              
        self.yelasticity = N * (T-1) * \
            (np.tile(np.average(zrhs, axis = 0)[:, None], (1, K)) /  
            np.tile(np.average(y, axis = 0)[None, :], (S, 1))) * \
            (np.reshape(np.sum((np.kron(pprob.T, np.ones(S)[:, None]) * 
            np.tile(np.average(x, axis = 0)[None,:]  ** 2, (S * K, 1))) * 
            (np.tile(params, (1, K)) - np.dot(np.tile(np.reshape(params, 
            (K, S)).T, (K, 1)), pprob.T)), axis = 1)[:, None], (K, S)).T)

# Import data (Note: Save as "Windows Comma Separated" before importing)
all = np.genfromtxt("""/Users/hendersonhl/Documents/PhD/IndependentStudy/Transformation/Data/PanelTRANS.csv""", delimiter = ',', names = True)

# Extract data
saDD = all[['dd11', 'dd12', 'dd13', 'dd14', 'dd15', 'dd16', 
'dd17', 'dd18', 'dd19', 'dd110', 'dd111', 'dd112', 'dd113', 
'dd114', 'dd115', 'dd116']]
saAC = all[['dd21', 'dd22', 'dd23', 'dd24', 'dd25', 'dd26', 
'dd27', 'dd28', 'dd29', 'dd210', 'dd211', 'dd212', 'dd213', 
'dd214', 'dd215', 'dd216']]
saz = all[['landprod','laborprod','whiredlabor','ppricelandimp',
'pfrijol','pbeef','pcafe','pmaiz','parroz','credit','title',
'techassist','agehead','educhead']]
salandop = all[['landop']]
sasbeancorn = all[['sbeancorn']]
saslivst = all[['slivst']]

# Copy the structured array into an array
DD = saDD.view((float, 16))
AC = saAC.view((float, 16))
z = saz.view((float, 14))
landop = salandop.view((float, 1))
sbeancorn = sasbeancorn.view((float, 1))
slivst = saslivst.view((float, 1))

# Plot landop data
plt.clf()
n, bins, patches = plt.hist(landop, bins=120, range=(0,80), normed=True, facecolor='grey')
plt.xlabel('Land Operated')
plt.ylabel('Probability')
plt.grid(True)
plt.minorticks_on()
plt.tick_params(axis='x', which='minor', direction='out', top='off')
plt.savefig('/Users/hendersonhl/Documents/PhD/Dissertation/All/TransLandOp.pdf')

# Plot sbeancorn data
plt.clf()
n, bins, patches = plt.hist(sbeancorn, bins=100, range=(0,1), normed=True, facecolor='grey')
plt.xlabel('Output Share (Beans and Maize)')
plt.ylabel('Probability')
plt.grid(True)
plt.minorticks_on()
plt.tick_params(axis='x', which='minor', direction='out', top='off')
tick_locs = ([10, 20, 30, 40])
tick_lbls = ([0.1, 0.2, 0.3, 0.4])
plt.yticks(tick_locs, tick_lbls)
plt.savefig('/Users/hendersonhl/Documents/PhD/Dissertation/All/TransSbeancorn.pdf')

# Plot slivst data
plt.clf()
n, bins, patches = plt.hist(slivst, bins=100, range=(0,1), normed=True, facecolor='grey')
plt.xlabel('Output Share (Livestock)')
plt.ylabel('Probability')
plt.grid(True)
plt.minorticks_on()
plt.tick_params(axis='x', which='minor', direction='out', top='off', right='off')
tick_locs = ([10, 20, 30])
tick_lbls = ([0.1, 0.2, 0.3])
plt.yticks(tick_locs, tick_lbls)
plt.savefig('/Users/hendersonhl/Documents/PhD/Dissertation/All/TransSlivst.pdf')

# Define prior on the TPM
ACprior = np.array([
[0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625],
[0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833, 1e-06,0.0833, 1e-06, 1e-06, 1e-06],
[0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769, 1e-06, 1e-06, 1e-06],
[0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833, 1e-06,0.0833,0.0833, 1e-06, 1e-06, 1e-06],
[0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714, 1e-06,0.0714, 1e-06],
[0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625],
[0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714, 1e-06,0.0714,0.0714, 1e-06,0.0714,0.0714],
[0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667, 1e-06],
[0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667, 1e-06],
[0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625],
[0.0714,0.0714, 1e-06,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714, 1e-06],
[0.0833, 1e-06, 1e-06,0.0833,0.0833,0.0833, 1e-06,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833, 1e-06,0.0833,0.0833],
[0.0714,0.0714, 1e-06,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714, 1e-06,0.0714,0.0714],
[0.0833, 1e-06, 1e-06, 1e-06, 1e-06,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833],
[0.1000, 1e-06, 1e-06, 1e-06, 1e-06,0.1000, 1e-06, 1e-06,0.1000,0.1000,0.1000,0.1000,0.1000,0.1000,0.1000,0.1000],
[0.0909, 1e-06, 1e-06, 1e-06, 1e-06,0.0909,0.0909,0.0909,0.0909,0.0909, 1e-06,0.0909,0.0909,0.0909,0.0909,0.0909]])

DDprior = np.array([
[0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625],
[0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769, 1e-06, 1e-06, 1e-06],
[0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769, 1e-06, 1e-06, 1e-06],
[0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833,0.0833, 1e-06,0.0833,0.0833, 1e-06, 1e-06, 1e-06],
[0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714, 1e-06,0.0714, 1e-06],
[0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667, 1e-06],
[0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625],
[0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667, 1e-06],
[0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667, 1e-06],
[0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625],
[0.0769,0.0769, 1e-06,0.0769,0.0769,0.0769,0.0769,0.0769,0.0769, 1e-06,0.0769,0.0769,0.0769,0.0769,0.0769, 1e-06],
[0.0714,0.0714,0.0714,0.0714,0.0714,0.0714, 1e-06,0.0714,0.0714,0.0714,0.0714,0.0714,0.0714, 1e-06,0.0714,0.0714],
[0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667,0.0667, 1e-06,0.0667,0.0667],
[0.0909, 1e-06, 1e-06, 1e-06,0.0909,0.0909,0.0909, 1e-06,0.0909, 1e-06,0.0909,0.0909,0.0909,0.0909,0.0909,0.0909],
[0.1000, 1e-06, 1e-06, 1e-06, 1e-06,0.1000, 1e-06, 1e-06,0.1000,0.1000,0.1000,0.1000,0.1000,0.1000,0.1000,0.1000],
[0.0909, 1e-06, 1e-06, 1e-06,0.0909,0.0909,0.0909,0.0909,0.0909, 1e-06, 1e-06,0.0909,0.0909,0.0909,0.0909,0.0909]])

# Model using AC data
Model = Markov(data = AC, periods = 3, z = z, pprior = ACprior, esupport = 3)
Model.fit(gtol = 1e-04)  
Model.diagnostics()
Model.pmf(Model.params) 
Model.yelasticity()
print('AC model')
print('Entropy ratio statistic:')
print(Model.entratio)
print('P-value:')
print(Model.pval)
print('Normalized entropy measure for TPM:')
print(Model.norment)
print('Normalized entropy measure for Kth row of TPM:')
print(Model.normentk)
print('Elasticities:')
print(Model.yelasticity)

print('TPM:')
print(Model.P)
#print('Standard errors:')
#Model.se(Model.params) 
#print(Model.se)

# Calculate standard errors of predicted values (see Greene [2008] pg. 99 for details)
np.set_printoptions(precision = 4, linewidth=160, suppress = True)
x1 = np.average(AC[:1208], axis=0)[None,:]*100
x2 = np.average(AC[1208:2416], axis=0)[None,:]*100
x3 = np.average(AC[2416:], axis=0)[None,:]*100
x = np.vstack((x1,x2))
x0 = x3
scalar = 1.0 + np.dot(np.dot(x0,np.linalg.pinv(np.dot(x.T, x))),x0.T)[0][0]
ypred = np.dot(x, Model.P)
yactual = np.vstack((x2, x3))
resid = yactual - ypred
resid_sq = resid**2
var = np.sum(resid_sq, axis=0)*0.5
var_pred = var*scalar
se_pred_AC = np.sqrt(var_pred)

# Model using DD data
Model = Markov(data = DD, periods = 3, z = z, pprior = DDprior, esupport = 3)
Model.fit(gtol = 1e-04)  
Model.diagnostics()
Model.pmf(Model.params) 
Model.yelasticity()
print('DD model')
print('Entropy ratio statistic:')
print(Model.entratio)
print('P-value:')
print(Model.pval)
print('Normalized entropy measure for TPM:')
print(Model.norment)
print('Normalized entropy measure for Kth row of TPM:')
print(Model.normentk)
print('Elasticities:')
print(Model.yelasticity)

print('TPM:')
print(Model.P)
#print('Standard errors:')
#Model.se(Model.params) 
#print(Model.se)

# Calculate standard errors of predicted values (see Greene [2008] pg. 99 for details)
x1 = np.average(DD[:1208], axis=0)[None,:]*100
x2 = np.average(DD[1208:2416], axis=0)[None,:]*100
x3 = np.average(DD[2416:], axis=0)[None,:]*100
x = np.vstack((x1,x2))
x0 = x3
scalar = 1.0 + np.dot(np.dot(x0,np.linalg.pinv(np.dot(x.T, x))),x0.T)[0][0]
ypred = np.dot(x, Model.P)
yactual = np.vstack((x2, x3))
resid = yactual - ypred
resid_sq = resid**2
var = np.sum(resid_sq, axis=0)*0.5
var_pred = var*scalar
se_pred_DD = np.sqrt(var_pred)
